/* Describe the available memory blocks */
MEMORY
{
  flash    : ORIGIN = 0x00002000, LENGTH = 504K
  ram      : ORIGIN = 0x10000000, LENGTH = 32K
  ram2     : ORIGIN = 0x2007C000, LENGTH = 16K
  ram3     : ORIGIN = 0x20080000, LENGTH = 16K
}

/* Stack grows downwards from the end of the first (main) RAM block.
 * Keep 32 bytes free for calls to the IAP flash programming functions.
 * Keep another 16 bytes free to work around a GDB issue where it reads
 * beyond the end of the stack.
 */
__stack_end__ = ORIGIN( ram ) + LENGTH( ram ) - 32 - 16;



/* --------------------------------------------------------
 * FreeRTOS -- Definitions for Memory Protection Unit (MPU)
 * --------------------------------------------------------
 */

/* The 'FLASH_segment' will be set to unprivileged read-only access.
 * Include the whole flash address range.
 */
__FLASH_segment_start__ = ORIGIN( flash );
__FLASH_segment_end__   = __FLASH_segment_start__ + LENGTH( flash );

/* The 'SRAM_segment' will be set to unprivileged read/write access.
 * Include the full range of the first RAM block. Access to the other
 * RAM blocks must be allowed on a per task basis!
 */
__SRAM_segment_start__ = ORIGIN( ram );
__SRAM_segment_end__   = __SRAM_segment_start__ + LENGTH( ram );

/* Define the size of the Flash (functions) and RAM (data) regions which are
 * protected for kernel use. The size must be sufficient for the kernel,
 * and it must be a valid MPU region size.
 * VERY IMPORTANT: Make sure that __FLASH_segment_start__ is aligned to
 *                 at least _sizeof_privileged_functions_!
 */
_sizeof_privileged_functions_ = 8K;
_sizeof_privileged_data_      = 512;

/* FreeRTOS assumes that the privileged functions always start at the
 * beginning of the FLASH_segment. We only have to define the end address.
 */
__privileged_functions_end__  = __FLASH_segment_start__ + _sizeof_privileged_functions_;

/* Protected RAM data for kernel use. Make sure to place it at the beginning
 * of the .data section below (We have to assume that this can be initialized
 * RAM, so it cannot be put in the .bss section).
 */
__privileged_data_start__ = ORIGIN( ram );
__privileged_data_end__   = __privileged_data_start__ + _sizeof_privileged_data_;





SECTIONS
{
    . = ORIGIN(flash);

    .text :
    {
        KEEP(*(.vectors))
/*    *(privileged_functions)/

    . = ALIGN(_sizeof_privileged_functions_);
*/
        *(.text .text.*)
        *(.rodata .rodata.*)
        *(.gnu.linkonce.t.*)
        *(.gnu.linkonce.r.*)
    } >flash

    /* Ordered library section goes separately. */
    .ARM.exidx :
    {
        *(.ARM.exidx* .gnu.linkonce.armexidx.*)
    } >flash

    __end_of_text__ = .;
    __text_size__ = . - ORIGIN(flash);

    .data :
    {
        __data_start__ = .;
/*    *(privileged_data)

    . = ALIGN(_sizeof_privileged_data_);
*/
        *(.data .data.*)
        __data_end__ = ALIGN(4);
    } >ram AT>flash
    __data_size__ = SIZEOF(.data);


    .noinit :
  {
    __noinit_start__ = .;
    *(.noinit)
    __noinit_end__ = ALIGN(4);
  } >ram
  __noinit_size__ = SIZEOF(.noinit);

  .bss :
  {
    __bss_start__ = .;
    *(.bss COMMON)
    __bss_end__ = ALIGN(4);
    __newlib_heap_start__ = .;
  } >ram
  __bss_size__ = SIZEOF(.bss);


    __heap_start = .;

  .ram2data :
  {
    __data2_start__ = .;
    *(.ram2data)
    __data2_end__ = ALIGN(4);
  } >ram2 AT>flash
  __data2_size__ = SIZEOF(.ram2data);

  .ram2bss (NOLOAD) :
  {
    __bss2_start__ = .;
    *(.ram2bss)
    __bss2_end__ = ALIGN(4);
  } >ram2
  __bss2_size__ = SIZEOF(.ram2bss);

  .ram3data :
  {
    __data3_start__ = .;
    *(.ram3data)
    __data3_end__ = ALIGN(4);
  } >ram3 AT>flash
  __data3_size__ = SIZEOF(.ram3data);

  .ram3bss (NOLOAD) :
  {
    __bss3_start__ = .;
    *(.ram3bss)
    __bss3_end__ = ALIGN(4);
  } >ram3
  __bss3_size__ = SIZEOF(.ram3bss);


    __ram_size__  = (__data_size__ + __noinit_size__ + __bss_size__);
    __ram2_size__ = (__data2_size__ + __bss2_size__);
    __ram3_size__ = (__data3_size__ + __bss3_size__);


    __flash_size__ = (__text_size__ + __data_size__ + __data2_size__ + __data3_size__);

/*
    .debug_info     0 : { *(.debug_info .gnu.linkonce.wi.*) }

    .debug_ranges   0 : { *(.debug_ranges) }
    .debug_aranges  0 : { *(.debug_aranges) }
*/
}

