
#include <stdio.h>
#include <string.h>

#include "lpclib.h"

#include "bsp-fifisdr.h"
#include "params.h"


#if defined(__ICCARM__)
    #define __stdout __iar_Stdout
#endif

#if defined(__GNUC__)
    #define __stdout (*stdout)
    #define _Handle _file
#endif

#if !defined(__ICCARM__)
struct __FILE {
    uint8_t _Handle;                                    /**< Dummy! */
};
#endif

volatile int32_t ITM_RxBuffer;                          /**< Declared in core_cm3.h */


I2C_Handle i2c_bus;
uint8_t si570Address;
SI570_Handle vco;

const I2C_Config i2c_bitrate400k_config = {
    .opcode = I2C_OPCODE_SET_BITRATE,
    { .bitrate = 400000ul, }
};



DMA_Handle gpdma;
#if LPCLIB_USB
USB_Handle usbHardware;
#endif


/** Probe all I2C addresses to find the Si570.
 *
 *  The Si570 is the only device on the bus. The first ACK response must be the Si570.
 *  Return address=0 if no response is received for any of the probed addresses.
 *
 *  \param[in] handle Bus handle
 *  \retval 0 No Si570 found
 *  \retval !=0 Address of Si570
 */
static uint8_t BSP_probeSi570 (I2C_Handle handle)
{
    uint8_t address;

    /* Probe valid address range.
     * NOTE: The Si570 may use addresses below 0x08 that are not allowed by the I2C spec!
     */
    for (address = 0x01; address < 0x78; address++) {
        if (I2C_probe(handle, address) == LPCLIB_SUCCESS) {
            return address;                             /* Device found! */
        }
    }

    return 0;                                           /* Bad luck */
}



/** DMA driver memory allocation handler. */
static void BSP_dmaMemoryAllocation (LPCLIB_Event event)
{
    uint32_t size;

    switch (event.opcode) {
    case DMA_EVENT_MEMALLOC:
        size = *((uint32_t *)event.parameter);
        *((void **)event.parameter) = pvPortMalloc(size); //TODO
        break;

    case DMA_EVENT_MEMFREE:
        vPortFree(*((void **)event.parameter));  //TODO
        break;

    default:
        event.parameter = NULL;
        break;
    }
}

static const DMA_Config dmaInstallHandler = {
    .opcode = DMA_OPCODE_SET_SYSTEM_CALLBACK,
    {.callback = BSP_dmaMemoryAllocation, },
};



#define NORMAL_PIN(func) \
    IOCON_makeConfigD(PIN_FUNCTION_##func, PIN_PULL_NONE, PIN_OPENDRAIN_OFF)
#define OPENDRAIN_PIN(func) \
    IOCON_makeConfigD(PIN_FUNCTION_##func, PIN_PULL_NONE, PIN_OPENDRAIN_ON)


void BSP_init (void)
{
    IOCON_open();
#if BOARD_TYPE == BOARD_KAI2
    IOCON_configurePin(PIN_P0_0,  OPENDRAIN_PIN(3));    /* SDA1 */
    IOCON_configurePin(PIN_P0_1,  OPENDRAIN_PIN(3));    /* SCL1 */
    IOCON_configurePin(PIN_P0_6,  IOCON_makeConfigD(    /* I2SRX_SDA */
                       PIN_FUNCTION_1, PIN_PULL_REPEATER, PIN_OPENDRAIN_OFF));
    IOCON_configurePin(PIN_P0_7,  NORMAL_PIN(1));       /* I2STX_CLK */
    IOCON_configurePin(PIN_P0_8,  NORMAL_PIN(1));       /* I2STX_WS */
    IOCON_configurePin(PIN_P0_22, NORMAL_PIN(0));       /* GPIO (ENAREG1) */
    IOCON_configurePin(PIN_P0_29, NORMAL_PIN(1));       /* USB_D+ */
    IOCON_configurePin(PIN_P0_30, NORMAL_PIN(1));       /* USB_D- */
    IOCON_configurePin(PIN_P1_4,  NORMAL_PIN(0));       /* GPIO (TEILER3) */
    IOCON_configurePin(PIN_P1_8,  NORMAL_PIN(0));       /* GPIO (TEILER2) */
    IOCON_configurePin(PIN_P1_9,  NORMAL_PIN(0));       /* GPIO (TEILER1) */
    IOCON_configurePin(PIN_P1_10, NORMAL_PIN(0));       /* GPIO (SI570_OE) */
    IOCON_configurePin(PIN_P1_18, NORMAL_PIN(0));       /* GPIO (LED) */
    IOCON_configurePin(PIN_P1_19, NORMAL_PIN(0));       /* GPIO (ENAREG2) */
    IOCON_configurePin(PIN_P1_30, NORMAL_PIN(2));       /* VBUS */
    IOCON_configurePin(PIN_P2_5,  NORMAL_PIN(0));       /* GPIO (PRESELECT1) */
    IOCON_configurePin(PIN_P2_6,  NORMAL_PIN(0));       /* GPIO (PRESELECT2) */
    IOCON_configurePin(PIN_P2_7,  NORMAL_PIN(0));       /* GPIO (PRESELECT3) */
    IOCON_configurePin(PIN_P2_8,  NORMAL_PIN(0));       /* GPIO (PRESELECT4) */
    IOCON_configurePin(PIN_P2_9,  NORMAL_PIN(1));       /* USB_CONNECT */
    IOCON_configurePin(PIN_P4_28, NORMAL_PIN(0));       /* GPIO (UDA1361_PWDN) */
    IOCON_configurePin(PIN_P4_29, NORMAL_PIN(1));       /* TX_MCLK */
#endif
#if BOARD_TYPE == BOARD_ROLF1
    IOCON_configurePin(PIN_P0_0,  OPENDRAIN_PIN(3));    /* SDA1 */
    IOCON_configurePin(PIN_P0_1,  OPENDRAIN_PIN(3));    /* SCL1 */
    IOCON_configurePin(PIN_P0_6,  IOCON_makeConfigD(    /* I2SRX_SDA */
                       PIN_FUNCTION_1, PIN_PULL_REPEATER, PIN_OPENDRAIN_OFF));
    IOCON_configurePin(PIN_P0_7,  NORMAL_PIN(1));       /* I2STX_CLK */
    IOCON_configurePin(PIN_P0_8,  NORMAL_PIN(1));       /* I2STX_WS */
    IOCON_configurePin(PIN_P0_9,  NORMAL_PIN(2));       /* MOSI1 */
    IOCON_configurePin(PIN_P0_9,  NORMAL_PIN(2));       /* MOSI1 */
    IOCON_configurePin(PIN_P0_11, NORMAL_PIN(0));       /* GPIO (SI570_OE) */
    IOCON_configurePin(PIN_P0_25, NORMAL_PIN(0));       /* GPIO (GUI_LIGHT) */
    IOCON_configurePin(PIN_P0_26, NORMAL_PIN(2));       /* AOUT */
    IOCON_configurePin(PIN_P0_29, NORMAL_PIN(1));       /* USB_D+ */
    IOCON_configurePin(PIN_P0_30, NORMAL_PIN(1));       /* USB_D- */
    IOCON_configurePin(PIN_P1_0,  NORMAL_PIN(0));       /* GPIO (GUI_SSEL) */
    IOCON_configurePin(PIN_P1_1,  NORMAL_PIN(0));       /* GPIO (GUI_...) */
    IOCON_configurePin(PIN_P1_4,  NORMAL_PIN(0));       /* GPIO (GUI_...) */
    IOCON_configurePin(PIN_P1_18, NORMAL_PIN(0));       /* GPIO (LED) */
    IOCON_configurePin(PIN_P1_30, NORMAL_PIN(2));       /* VBUS */
    IOCON_configurePin(PIN_P1_31, NORMAL_PIN(2));       /* SCK1 */
    IOCON_configurePin(PIN_P2_4,  NORMAL_PIN(0));       /* GPIO (TEILER1) */
    IOCON_configurePin(PIN_P2_5,  NORMAL_PIN(0));       /* GPIO (TEILER2) */
    IOCON_configurePin(PIN_P2_6,  NORMAL_PIN(0));       /* GPIO (ENAREG1) */
    IOCON_configurePin(PIN_P2_7,  NORMAL_PIN(0));       /* GPIO (ENAREG2) */
    IOCON_configurePin(PIN_P2_9,  NORMAL_PIN(1));       /* USB_CONNECT */
    IOCON_configurePin(PIN_P4_28, NORMAL_PIN(0));       /* GPIO (UDA1361_PWDN) */
    IOCON_configurePin(PIN_P4_29, NORMAL_PIN(1));       /* TX_MCLK */
#endif
#if BOARD_TYPE == BOARD_ROLF2
    IOCON_configurePin(PIN_P0_0,  OPENDRAIN_PIN(3));    /* SDA1 */
    IOCON_configurePin(PIN_P0_1,  OPENDRAIN_PIN(3));    /* SCL1 */
    IOCON_configurePin(PIN_P0_2,  NORMAL_PIN(0));       /* GPIO (LED) */
    IOCON_configurePin(PIN_P0_6,  IOCON_makeConfigD(    /* I2SRX_SDA */
                       PIN_FUNCTION_1, PIN_PULL_REPEATER, PIN_OPENDRAIN_OFF));
    IOCON_configurePin(PIN_P0_7,  NORMAL_PIN(1));       /* I2STX_CLK */
    IOCON_configurePin(PIN_P0_8,  NORMAL_PIN(1));       /* I2STX_WS */
    IOCON_configurePin(PIN_P0_10, NORMAL_PIN(0));       /* GPIO (SI570_OE) */
    IOCON_configurePin(PIN_P0_11, NORMAL_PIN(0));       /* GPIO (ENAREG1) */
    IOCON_configurePin(PIN_P0_16, NORMAL_PIN(0));       /* GPIO (PLL2) */
    IOCON_configurePin(PIN_P0_22, NORMAL_PIN(0));       /* GPIO (PRESEL5) */
    IOCON_configurePin(PIN_P0_29, NORMAL_PIN(1));       /* USB_D+ */
    IOCON_configurePin(PIN_P0_30, NORMAL_PIN(1));       /* USB_D- */
    IOCON_configurePin(PIN_P1_8,  NORMAL_PIN(0));       /* GPIO (RESET_AD1974) */
    IOCON_configurePin(PIN_P1_18, NORMAL_PIN(0));       /* GPIO (PRESEL3) */
    IOCON_configurePin(PIN_P1_20, NORMAL_PIN(0));       /* GPIO (CCLK) */
    IOCON_configurePin(PIN_P1_22, NORMAL_PIN(0));       /* GPIO (PRESEL4) */
    IOCON_configurePin(PIN_P1_24, NORMAL_PIN(0));       /* GPIO (CIN) */
    IOCON_configurePin(PIN_P1_25, NORMAL_PIN(0));       /* GPIO (PRESEL1) */
    IOCON_configurePin(PIN_P1_26, NORMAL_PIN(0));       /* GPIO (ENAREG2) */
    IOCON_configurePin(PIN_P1_30, NORMAL_PIN(2));       /* VBUS */
    IOCON_configurePin(PIN_P2_0,  NORMAL_PIN(0));       /* GPIO (INH7046) */
    IOCON_configurePin(PIN_P2_1,  NORMAL_PIN(1));       /* GPIO (PWM1.2) */
    IOCON_configurePin(PIN_P2_2,  NORMAL_PIN(0));       /* GPIO (CLATCH) */
    IOCON_configurePin(PIN_P2_6,  NORMAL_PIN(0));       /* GPIO (TEILER1) */
    IOCON_configurePin(PIN_P2_7,  NORMAL_PIN(0));       /* GPIO (TEILER2) */
    IOCON_configurePin(PIN_P2_8,  NORMAL_PIN(0));       /* GPIO (PLL1) */
    IOCON_configurePin(PIN_P2_9,  NORMAL_PIN(1));       /* USB_CONNECT */
#endif
    if (IOCON_checkErrors() != LPCLIB_SUCCESS) {
        while (1);  //DEBUGGING ONLY
    }

    GPIO_open();
#if BOARD_TYPE == BOARD_KAI2
    GPIO_setDirBit(LED, ENABLE);            GPIO_writeBit(LED,0);
    GPIO_setDirBit(ENAREG1, ENABLE);        GPIO_writeBit(ENAREG1,1);
    GPIO_setDirBit(ENAREG2, ENABLE);        GPIO_writeBit(ENAREG2,1);
    GPIO_setDirBit(UDA1361_PWDN, ENABLE);   GPIO_writeBit(UDA1361_PWDN,0);
    GPIO_setDirBit(TEILER1, ENABLE);
    GPIO_setDirBit(TEILER2, ENABLE);
    GPIO_setDirBit(TEILER3, ENABLE);
    GPIO_setDirBit(PRESELECT1, ENABLE);
    GPIO_setDirBit(PRESELECT2, ENABLE);
    GPIO_setDirBit(PRESELECT3, ENABLE);
    GPIO_setDirBit(PRESELECT4, ENABLE);     GPIO_writeBit(PRESELECT4, 0);
    GPIO_setDirBit(SI570_OE, ENABLE);       GPIO_writeBit(SI570_OE, 1);
#endif
#if BOARD_TYPE == BOARD_ROLF1
    GPIO_setDirBit(LED, ENABLE);            GPIO_writeBit(LED,0);
    GPIO_setDirBit(ENAREG1, ENABLE);        GPIO_writeBit(ENAREG1,1);
    GPIO_setDirBit(ENAREG2, ENABLE);        GPIO_writeBit(ENAREG2,1);
    GPIO_setDirBit(UDA1361_PWDN, ENABLE);   GPIO_writeBit(UDA1361_PWDN,0);
    GPIO_setDirBit(TEILER1, ENABLE);
    GPIO_setDirBit(TEILER2, ENABLE);
    GPIO_setDirBit(SI570_OE, ENABLE);       GPIO_writeBit(SI570_OE, 1);
#endif
#if BOARD_TYPE == BOARD_ROLF2
    GPIO_setDirBit(LED, ENABLE);            GPIO_writeBit(LED,0);
    GPIO_setDirBit(ENAREG1, ENABLE);        GPIO_writeBit(ENAREG1,1);
    GPIO_setDirBit(ENAREG2, ENABLE);        GPIO_writeBit(ENAREG2,1);
    GPIO_setDirBit(PRESELECT1, ENABLE);
    GPIO_setDirBit(PRESELECT3, ENABLE);
    GPIO_setDirBit(PRESELECT4, ENABLE);
    GPIO_setDirBit(PRESELECT5, ENABLE);
    GPIO_setDirBit(PLL1, ENABLE);
    GPIO_setDirBit(PLL2, ENABLE);
    GPIO_setDirBit(CIN, ENABLE);
    GPIO_setDirBit(CCLK, ENABLE);
    GPIO_setDirBit(RESET_AD1974, ENABLE);
    GPIO_setDirBit(CLATCH, ENABLE);         GPIO_writeBit(CLATCH, 1);
    GPIO_setDirBit(INH7046, ENABLE);
    GPIO_setDirBit(TEILER1, ENABLE);
    GPIO_setDirBit(TEILER2, ENABLE);
    GPIO_setDirBit(SI570_OE, ENABLE);       GPIO_writeBit(SI570_OE, 1);
#endif

    DMA_open(DMA0, &gpdma);
    DMA_ioctl(gpdma, &dmaInstallHandler);

    I2C_open(I2C1, &i2c_bus);
    I2C_ioctl(i2c_bus, &i2c_bitrate400k_config);
    si570Address = BSP_probeSi570(i2c_bus);             /* Identify slave address of Si570 */
    SI570_open(i2c_bus, si570Address, &vco);

    CLKPWR_setUsbClock();
    USB_open(USB0, &usbHardware);
}


/********** Standard Library I/O **********/

#if (defined(__GNUC__))

#include <sys/stat.h>

extern uint32_t __heap_start;


int _fstat (int fd, struct stat *st);
int _fstat (int fd, struct stat *st)
{
    (void) fd;
    (void) st;

    return -1;
}


int _read (int fd, char *ptr, int len);
int _read (int fd, char *ptr, int len)
{
    (void) fd;

    int bytesUnRead = len;
    int32_t c;

    c = ITM_ReceiveChar();
    if (c >= 0) {
        *ptr = c;
        --bytesUnRead;
    }

    return len - bytesUnRead;
}


int _write (int fd, const char *ptr, int len);
int _write (int fd, const char *ptr, int len)
{
    (void) fd;

    int bytesUnWritten = -1;

#if 1
    int n = 0;
    while (++n < 10000) {
        if (*((volatile uint8_t *)0x20083800) == 0) {
            if (len > 252) {
                len = 252; //TODO
            }
            memcpy((void *)0x20083804, ptr, len);
            *((volatile uint8_t *)0x20083800) = len;

            break;
        }
    }

    bytesUnWritten = 0;
#else
    int i;
    char c;

    for (i = 0; i < len; i++) {
        ITM_SendChar(*ptr);

        if (*ptr == '\n') {
            c = '\r';
            ITM_SendChar(c);
        }

        ++ptr;
    }
    bytesUnWritten = len - i;
#endif

    return len - bytesUnWritten;
}


int _close (int fd);
int _close (int fd)
{
    (void) fd;

    return 0;
}


int _lseek (int fd, int ptr, int dir);
int _lseek (int fd, int ptr, int dir)
{
    (void)fd;
    (void)ptr;
    (void)dir;

    return 0;
}


int _isatty (int fd);
int _isatty (int fd)
{
    return (fd <= 2) ? 1 : 0;   /* one of stdin, stdout, stderr */
}


caddr_t _sbrk (int incr);
caddr_t _sbrk (int incr)
{
    static caddr_t heap = NULL;
    caddr_t prev_heap;

    if (heap == NULL) {
        heap = (caddr_t)&__heap_start;
    }

    prev_heap = heap;
    heap += incr;

    return prev_heap;
}


#else


FILE __stdout;
FILE __stdin;


int fputc (int ch, FILE *f)
{
    (void) f;

    if (ch == '\n') {
        ch = '\r';
    }
    ITM_SendChar(ch);

    return ch;
}


#if defined(__ICCARM__)
int fgetc (FILE *f)
#else
int getc (FILE *f)
#endif
{
    int c;

    c = ITM_ReceiveChar();
    if (c >= 0) {
        return c;
    }

    return EOF;
}

int ferror (FILE *f)
{
    return EOF;
}
#endif

