
#include <stdio.h>

#include "lpclib.h"

#include "task-sys.h"
#include "task-gui.h"
#include "usbuser.h"

#include "system_LPC17xx.h"
#include "bsp-fifisdr.h"
#include "params.h"


/* Task priorities. */
#define mainINIT_TASK_PRIORITY        osPriorityHigh
#define mainGUI_TASK_PRIORITY         osPriorityLow
#define mainSYS_TASK_PRIORITY         osPriorityNormal
#define mainUSB_TASK_PRIORITY         osPriorityAboveNormal
#define mainAUDIO_TASK_PRIORITY       osPriorityRealtime

osThreadDef(GUI_task, mainGUI_TASK_PRIORITY, 1, 1*configMINIMAL_STACK_SIZE*4);
osThreadDef(SYS_task, mainSYS_TASK_PRIORITY, 1, 4*configMINIMAL_STACK_SIZE*4);
osThreadDef(USBUSER_task, mainUSB_TASK_PRIORITY, 1, 4*configMINIMAL_STACK_SIZE*4);
osThreadDef(FIFIAUDIO_task, mainAUDIO_TASK_PRIORITY, 1, 6*configMINIMAL_STACK_SIZE*4);


/** Task to initialize the system and all other tasks.
 *
 *  Initializes the system already under RTOS control. This is done because many drivers expect
 *  semaphores/mutexes to work. Sets up other tasks, then terminates itself after all initialization
 *  has been done.
 */
static void systemInit_task (void const *pParameters)
{
    (void) pParameters;
    int n;

#if defined (__GNUC__)
    /* TODO
     * The 'manual' call to __sinit() is necessary with NEWLIB for the following reasons:
     * NEWLIB automatically compiles with option _REENT_SMALL if the compiler uses Thumb-2
     * (true for Cortex-M3). In this case the standard file (stdout etc) pointers in the global
     * impure_data point to shortened fake FILE* in ROM, until they get initialized by __sinit().
     * It is unclear who is responsible for calling __sinit(), but doing it here solves it for now.
     */
    __sinit(_global_impure_ptr);
#endif

    /* Priority of peripheral interrupts depends on the application, and therefore the driver
     * library does not set the priority.
     * After reset, all interrupts have highest priority, and this can lead to nasty failures
     * if your application uses an RTOS. Therefore it is a good idea to preset all interrupt
     * priorities to a safe default (the lowest priority). You may afterwards raise the
     * priority of individual interrupts as required by the application.
     */
    for (n = 0; n < 240; n++) {                         /* 240 is the maximum for any M3 implementation */
        NVIC_SetPriority((IRQn_Type)n, (1u << __NVIC_PRIO_BITS) - 1);
    }

    NVIC_SetPriority(I2C1_IRQn, configI2C1_INTERRUPT_PRIORITY);
    NVIC_SetPriority(USB_IRQn, configUSB_INTERRUPT_PRIORITY);
    NVIC_SetPriority(DMA_IRQn, configDMA_INTERRUPT_PRIORITY);
    NVIC_SetPriority(I2S_IRQn, configI2S_INTERRUPT_PRIORITY);

    paramsRead();                                       /* Global system settings */

    /* Prepare tasks */
    SYS_initModule();
    GUI_initModule();
    USBUSER_initModule();
    FIFIAUDIO_initModule();

    BSP_init();

    /* Create tasks */
    osThreadCreate(osThread(GUI_task), NULL);
    osThreadCreate(osThread(SYS_task), NULL);
    osThreadCreate(osThread(USBUSER_task), NULL);
    osThreadCreate(osThread(FIFIAUDIO_task), NULL);

    /* The systemInit task is no longer needed */
    osThreadTerminate(osThreadGetId());
}

osThreadDef(systemInit_task, mainINIT_TASK_PRIORITY, 1, 9*configMINIMAL_STACK_SIZE*4);


int main (void)
{
    /* Start the scheduler. */
    osKernelStart(osThread(systemInit_task), NULL);
    while(1)
        ;
}


/* Override system initialization (runs before main()!).
 * Set CPU clock to speed up the startup process.
 */
void SystemInit (void)
{
    /* Erratum PCLKSELx.1: Must set dividers before enabling PLL0! */
    CLKPWR_setDivider(CLKPWR_DIVIDER_TIMER3, CLKPWR_RATIO_4);   /* used in fifiaudio.c */
    CLKPWR_setDivider(CLKPWR_DIVIDER_I2S, CLKPWR_RATIO_2);      /* NOTE: LPC17xx erratum:
                                                                 * PCLK(I2S) must be 72 MHz max!
                                                                 */

    SystemCoreClock = 0;
    CLKPWR_setCpuClock(configCPU_CLOCK_HZ);             /* Take frequency from FreeRTOSConfig.h */
}


#if 1
/* FreeRTOS stubs */
extern void vApplicationIdleHook (void);
void vApplicationIdleHook (void)
{
}

void vApplicationStackOverflowHook (xTaskHandle *pxTask, signed portCHAR *pcTaskName);
void vApplicationStackOverflowHook (xTaskHandle *pxTask, signed portCHAR *pcTaskName)
{
    (void) pxTask;
    (void) pcTaskName;

}
#endif
