/* Copyright (c) 2010, DF9DQ
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this list of conditions
 * and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions
 * and the following disclaimer in the documentation and/or other materials provided with the
 * distribution.
 * Neither the name of the author nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <stdlib.h>

#include "lpclib.h"
#include "bsp-fifisdr.h"

#include "usbuser.h"



/** Enumerate all string descriptors.
 *
 *  Use these names in the string descriptor table and the device/configuration descriptors
 *  to achieve an automatic numbering of all string descriptors.
 *
 *  The first entry must be assigned index 1.
 *  The last entry must be USB_NUM_STRING_DESCRIPTORS_PLUS1.
 */
enum {
    USBSTR_IDVENDOR = 1,
    USBSTR_IDPRODUCT,
    USBSTR_SERIALNUMBER,
    USBSTR_SOUNDCARD,
    USBSTR_STREAM1,
    USBSTR_SOUNDCARD2,
    USBSTR_STREAM2,
    USBSTR_SOFTROCK,
    USBSTR_CHANNEL_I, USBSTR_CHANNEL_Q,         /* MUST have consecutive numbers! */
    USBSTR_CHANNEL_USB, USBSTR_CHANNEL_LSB,     /* MUST have consecutive numbers! */
    USBSTR_PREAMP_UDA,
    USBSTR_COMPRESSOR,
    USBSTR_UDA_INPUT_TERMINAL,
    USBSTR_DEMOD_INPUT_TERMINAL,

    USB_NUM_STRING_DESCRIPTORS_PLUS1
};


#define USB_NUM_LANGUAGES                   2

static DECLARE_LANGUAGE_ID_DESCRIPTOR(fifisdr_languages,
                                      USB_NUM_LANGUAGES,
                                      0x0407,   /* German           */
                                      0x0409,   /* English          */
                                      );

static DECLARE_STRING_DESCRIPTOR(str_idVendor_en, L"www.ov-lennestadt.de");
static DECLARE_STRING_DESCRIPTOR(str_idProduct_en, L"FiFi-SDR");
static DECLARE_STRING_DESCRIPTOR(str_soundcard_en, L"FiFi-SDR Soundcard");
static DECLARE_STRING_DESCRIPTOR(str_stream1_de, L"UDA1361 Eingang");
static DECLARE_STRING_DESCRIPTOR(str_stream1_en, L"UDA1361 Input");
static DECLARE_STRING_DESCRIPTOR(str_audiocontrol2_en, L"FiFi-SDR DSP");
static DECLARE_STRING_DESCRIPTOR(str_stream2_de, L"SSB");
static DECLARE_STRING_DESCRIPTOR(str_stream2_en, L"SSB");
static DECLARE_STRING_DESCRIPTOR(str_softrock_en, L"Softrock SDR");
static DECLARE_STRING_DESCRIPTOR(str_soundchannel_i_en, L"I");
static DECLARE_STRING_DESCRIPTOR(str_soundchannel_q_en, L"Q");
static DECLARE_STRING_DESCRIPTOR(str_soundchannel_usb_en, L"USB");
static DECLARE_STRING_DESCRIPTOR(str_soundchannel_lsb_en, L"LSB");
static DECLARE_STRING_DESCRIPTOR(str_preamp_uda_de, L"VV");
static DECLARE_STRING_DESCRIPTOR(str_preamp_uda_en, L"Preamp");
static DECLARE_STRING_DESCRIPTOR(str_uda_input_de, L"Basisband (I/Q)");
static DECLARE_STRING_DESCRIPTOR(str_uda_input_en, L"Complex Baseband (I/Q)");
static DECLARE_STRING_DESCRIPTOR(str_demod_input_de, L"DSP");
static DECLARE_STRING_DESCRIPTOR(str_demod_input_en, L"DSP");


/* The serial number is a special string in RAM that will be filled in at initialization time.
 * It holds a 64 character string with the serial number of the LPC device.
 * The string is initialized to a safe default.
 */
static __PACKED(struct {
    uint8_t     bLength;
    uint8_t     bDescriptorType;
    wchar_t     text[32];
}) str_iSerialNumber = {
    .bLength = 4,
    .bDescriptorType = USBD_STRING,
    .text = {L'*',},
};



static const struct {
    uint8_t id;
    const void *desc[USB_NUM_LANGUAGES];
} fifisdr_stringDescriptors[USB_NUM_STRING_DESCRIPTORS_PLUS1-1] =
{
    {.id = USBSTR_IDVENDOR,
      .desc = {&str_idVendor_en,
               &str_idVendor_en,}
    },
    { .id = USBSTR_IDPRODUCT,
      .desc = {&str_idProduct_en,
               &str_idProduct_en,}
    },
    { .id = USBSTR_SERIALNUMBER,
      .desc = {&str_iSerialNumber,
               &str_iSerialNumber},
    },
    { .id = USBSTR_SOUNDCARD,
      .desc = {&str_soundcard_en,
               &str_soundcard_en},
    },
    { .id = USBSTR_STREAM1,
      .desc[0] = &str_stream1_de,
      .desc[1] = &str_stream1_en,
    },
    { .id = USBSTR_SOUNDCARD2,
      .desc[0] = &str_audiocontrol2_en,
      .desc[1] = &str_audiocontrol2_en,
    },
    { .id = USBSTR_STREAM2,
      .desc[0] = &str_stream2_de,
      .desc[1] = &str_stream2_en,
    },
    { .id = USBSTR_SOFTROCK,
      .desc[0] = &str_softrock_en,
      .desc[1] = &str_softrock_en,
    },
    { .id = USBSTR_CHANNEL_I,
      .desc[0] = &str_soundchannel_i_en,
      .desc[1] = &str_soundchannel_i_en,
    },
    { .id = USBSTR_CHANNEL_Q,
      .desc[0] = &str_soundchannel_q_en,
      .desc[1] = &str_soundchannel_q_en,
    },
    { .id = USBSTR_CHANNEL_USB,
      .desc[0] = &str_soundchannel_usb_en,
      .desc[1] = &str_soundchannel_usb_en,
    },
    { .id = USBSTR_CHANNEL_LSB,
      .desc[0] = &str_soundchannel_lsb_en,
      .desc[1] = &str_soundchannel_lsb_en,
    },
    { .id = USBSTR_PREAMP_UDA,
      .desc[0] = &str_preamp_uda_de,
      .desc[1] = &str_preamp_uda_en,
    },
    { .id = USBSTR_UDA_INPUT_TERMINAL,
      .desc[0] = &str_uda_input_de,
      .desc[1] = &str_uda_input_en,
    },
    { .id = USBSTR_DEMOD_INPUT_TERMINAL,
      .desc[0] = &str_demod_input_de,
      .desc[1] = &str_demod_input_en,
    },
};



/** Return pointer to string descriptor of given index and language.
 */
static LPCLIB_Result fifisdr_getString (uint8_t index,
                                        uint16_t languageId,
                                        const void **pString)
{
    uint16_t theLanguage;
    uint16_t i;


    /* Special handling for index zero. This is the descriptor which specifies the ID codes
     * of the supported languages.
     */
    if (index == 0) {
        *pString = (void *)&fifisdr_languages;

        return LPCLIB_SUCCESS;
    }

    /* Determine the language index in the string table. Fall back to index zero (first language)
     * if an unknown language ID is specified.
     */
    theLanguage = 0;
    for (i = 0; i < USB_NUM_LANGUAGES; i++) {
        if (fifisdr_languages.langId[i] == languageId) {
            theLanguage = i;
            break;
        }
    }

    /* Search the list */
    for (i = 0; i < USB_NUM_STRING_DESCRIPTORS_PLUS1 - 1; i++) {
        /* Does the specified index match the string ID? */
        if (fifisdr_stringDescriptors[i].id == index) {
            *pString = fifisdr_stringDescriptors[i].desc[theLanguage];

            return LPCLIB_SUCCESS;
        }
    }

    return LPCLIB_ILLEGAL_PARAMETER;
}



/** FiFi-SDR device descriptor. */
static const USB_DeviceDescriptor fifisdr_deviceDescriptor = {
    .bLength                    = USBD_SIZE_DEVICE,
    .bDescriptorType            = USBD_DEVICE,
    .bcdUSB                     = 0x0110,
    .bDeviceClass               = USBC_BASE,
    .bDeviceSubClass            = 0,
    .bDeviceProtocol            = 0,
    .bMaxPacketSize0            = 64,
    .idVendor                   = 0x16C0,
    .idProduct                  = 0x05DC,
    .bcdDevice                  = 0x0100,
    .iManufacturer              = USBSTR_IDVENDOR,
    .iProduct                   = USBSTR_IDPRODUCT,
    .iSerialNumber              = USBSTR_SERIALNUMBER,
    .bNumConfigurations         = USBCONFIG_NUM_CONFIGURATIONS,
};


/** FiFi-SDR configuration 1 */
static const __PACKED(struct {
    USB_ConfigurationDescriptor                     config;
    USB_InterfaceDescriptor                         softrockInterface;

#if USBCONFIG_SOUNDCARD
    USB_InterfaceDescriptor                         audioControlInterface1;
    USBAUDIO_ControlInterfaceDescriptor             audioControlInterfaceHeader1;
    uint8_t                                         audioControlInterfaceHeader1_baInterfaceNr[1];
    USBAUDIO_InputTerminalDescriptor                audioUdaInputTerminal;
    USBAUDIO_OutputTerminalDescriptor               audioOutputTerminalIq;
    USBAUDIO_FeatureUnitDescriptor                  audioFeatureUnitUda;
    uint8_t                                         audioFeatureUnitUda_bmaControls[3];
    uint8_t                                         audioFeatureUnitUda_iFeature;
    USB_InterfaceDescriptor                         audioStreamingInterface_1_0;
    USB_InterfaceDescriptor                         audioStreamingInterface_1_1;
    USBAUDIO_StreamingInterfaceDescriptor           audioClassStreamingInterface1;
    USBAUDIO_MicrophoneType1FormatTypeDescriptor    audioFormatType1;
    uint8_t                                         audioSampleFrequencies1[2][3];
    USB_EndpointDescriptorIso                       audioEndpoint1;
    USBAUDIO_IsoDataEndpointDescriptor              audioDataEndpoint1;
    USB_InterfaceDescriptor                         audioStreamingInterface_1_2;
    USBAUDIO_StreamingInterfaceDescriptor           audioClassStreamingInterface2;
    USBAUDIO_MicrophoneType1FormatTypeDescriptor    audioFormatType2;
#if BOARD_TYPE == BOARD_ROLF2
    uint8_t                                         audioSampleFrequencies2[3][3];
#else
    uint8_t                                         audioSampleFrequencies2[2][3];
#endif
    USB_EndpointDescriptorIso                       audioEndpoint2;
    USBAUDIO_IsoDataEndpointDescriptor              audioDataEndpoint2;

  #if USBCONFIG_DSP
    USB_InterfaceDescriptor                         audioDspControlInterface;
    USBAUDIO_ControlInterfaceDescriptor             audioDspControlInterfaceHeader;
    uint8_t                                         audioDspControlInterfaceHeader_baInterfaceNr[1];
    USBAUDIO_InputTerminalDescriptor                audioDspInputTerminal;
    USBAUDIO_OutputTerminalDescriptor               audioDspOutputTerminal;
    USBAUDIO_FeatureUnitDescriptor                  audioDspFeatureUnit;
    uint8_t                                         audioDspFeatureUnit_bmaControls[2];
    uint8_t                                         audioDspFeatureUnit_iFeature;
    USB_InterfaceDescriptor                         audioDspStreamingInterface_0;
    USB_InterfaceDescriptor                         audioDspStreamingInterface_1;
    USBAUDIO_StreamingInterfaceDescriptor           audioDspClassStreamingInterface;
    USBAUDIO_MicrophoneType1FormatTypeDescriptor    audioDspFormatType;
    uint8_t                                         audioDspSampleFrequencies[1][3];
    USB_EndpointDescriptorIso                       audioDspEndpoint;
    USBAUDIO_IsoDataEndpointDescriptor              audioDspDataEndpoint;
  #endif
#endif  /* #if USBCONFIG_SOUNDCARD */
}) fifisdr_configuration1 = {

    .config = {
        .bLength                = USBD_SIZE_CONFIGURATION,
        .bDescriptorType        = USBD_CONFIGURATION,
        .wTotalLength           = sizeof(fifisdr_configuration1),
        .bNumInterfaces         = USBCONFIG_NUM_INTERFACES,
        .bConfigurationValue    = 1,
        .iConfiguration         = 0,
        .bmAttributes           = 0x80,     /* bus-powered */
        .bMaxPower              = 250/2,
    },

    /* Softrock-40 Interface */
    .softrockInterface = {
        .bLength                = USBD_SIZE_INTERFACE,
        .bDescriptorType        = USBD_INTERFACE,
        .bInterfaceNumber       = USBCONFIG_INTERFACE_SOFTROCK_1,
        .bAlternateSetting      = 0,
        .bNumEndpoints          = 0,
        .bInterfaceClass        = USBC_VENDOR_SPECIFIC,
        .bInterfaceSubClass     = 0,
        .bInterfaceProtocol     = 0,
        .iInterface             = USBSTR_SOFTROCK,
    },

#if USBCONFIG_SOUNDCARD
    /* Standard AudioControl Interface */
    .audioControlInterface1 = {
        .bLength                = USBD_SIZE_INTERFACE,
        .bDescriptorType        = USBD_INTERFACE,
        .bInterfaceNumber       = USBCONFIG_INTERFACE_AUDIO_CONTROL_1,
        .bAlternateSetting      = 0,
        .bNumEndpoints          = 0,
        .bInterfaceClass        = USBC_AUDIO,
        .bInterfaceSubClass     = USBCS_AUDIO_AUDIOCONTROL,
        .bInterfaceProtocol     = 0,
        .iInterface             = USBSTR_SOUNDCARD,
    },

    /* Class-Specific Audio Control Interface */
    .audioControlInterfaceHeader1 = {
        .bLength                = USBD_SIZE_AUDIO_ST_HEADER(1),
        .bDescriptorType        = USBD_CS_INTERFACE,
        .bDescriptorSubtype     = USBD_AUDIO_ST_HEADER,
        .bcdADC                 = 0x0100,   /* 1.0 */
        .wTotalLength           =
                                  USBD_SIZE_AUDIO_ST_HEADER(1) +
                                  USBD_SIZE_AUDIO_ST_INPUT_TERMINAL +
                                  USBD_SIZE_AUDIO_ST_OUTPUT_TERMINAL +
                                  USBD_SIZE_AUDIO_ST_FEATURE_UNIT(2,1) +
                                  0,
        .bInCollection          = 1,
    },
    .audioControlInterfaceHeader1_baInterfaceNr = {
        USBCONFIG_INTERFACE_AUDIO_STREAMING_1,
    },

    /* Input Terminal (I/Q) */
    .audioUdaInputTerminal = {
        .bLength                = USBD_SIZE_AUDIO_ST_INPUT_TERMINAL,
        .bDescriptorType        = USBD_CS_INTERFACE,
        .bDescriptorSubtype     = USBD_AUDIO_ST_INPUT_TERMINAL,
        .bTerminalID            = USBCONFIG_UNIT_TERMINAL_IN_IQ,
        .wTerminalType          = 0x0201,
        .bAssocTerminal         = 0,        /* no association */
        .bNrChannels            = 2,
        .wChannelConfig         = 0x0003,
        .iChannelNames          = 0,
        .iTerminal              = USBSTR_UDA_INPUT_TERMINAL,
    },

    /* Output Terminal I/Q */
    .audioOutputTerminalIq = {
        .bLength                = USBD_SIZE_AUDIO_ST_OUTPUT_TERMINAL,
        .bDescriptorType        = USBD_CS_INTERFACE,
        .bDescriptorSubtype     = USBD_AUDIO_ST_OUTPUT_TERMINAL,
        .bTerminalID            = USBCONFIG_UNIT_TERMINAL_OUT_IQ,
        .wTerminalType          = 0x0101,   /* USB Streaming */
        .bAssocTerminal         = 0,        /* no association */
        .bSourceID              = USBCONFIG_UNIT_FEATURE_IQ,
        .iTerminal              = 0,
    },

    /* Feature Unit (UDA) */
    .audioFeatureUnitUda = {
        .bLength                = USBD_SIZE_AUDIO_ST_FEATURE_UNIT(2,1),
        .bDescriptorType        = USBD_CS_INTERFACE,
        .bDescriptorSubtype     = USBD_AUDIO_ST_FEATURE_UNIT,
        .bUnitID                = USBCONFIG_UNIT_FEATURE_IQ,
        .bSourceID              = USBCONFIG_UNIT_TERMINAL_IN_IQ,
        .bControlSize           = 1,        /* n = 1 Byte/channel */
    },
    .audioFeatureUnitUda_bmaControls = {
            0x03,               /* bmaControls(0)       Master: Mute+Volume */
            0x00,               /* bmaControls(1)                           */
            0x00,               /* bmaControls(2)                           */
    },
    .audioFeatureUnitUda_iFeature  = USBSTR_PREAMP_UDA,

    /* Standard AudioStreaming Interface 1 (alternate setting #0) */
    .audioStreamingInterface_1_0 = {
        .bLength                = USBD_SIZE_INTERFACE,
        .bDescriptorType        = USBD_INTERFACE,
        .bInterfaceNumber       = USBCONFIG_INTERFACE_AUDIO_STREAMING_1,
        .bAlternateSetting      = 0,
        .bNumEndpoints          = 0,
        .bInterfaceClass        = USBC_AUDIO,
        .bInterfaceSubClass     = USBCS_AUDIO_AUDIOSTREAMING,
        .bInterfaceProtocol     = 0,
        .iInterface             = USBSTR_STREAM1,
    },

    /* Standard AudioStreaming Interface 1 (alternate setting #1) */
    .audioStreamingInterface_1_1 = {
        .bLength                = USBD_SIZE_INTERFACE,
        .bDescriptorType        = USBD_INTERFACE,
        .bInterfaceNumber       = USBCONFIG_INTERFACE_AUDIO_STREAMING_1,
        .bAlternateSetting      = 1,
        .bNumEndpoints          = 1,
        .bInterfaceClass        = USBC_AUDIO,
        .bInterfaceSubClass     = USBCS_AUDIO_AUDIOSTREAMING,
        .bInterfaceProtocol     = 0,
        .iInterface             = USBSTR_STREAM1,
    },

    /* Audio Stream Audio Class */
    .audioClassStreamingInterface1 = {
        .bLength                = 7,
        .bDescriptorType        = USBD_CS_INTERFACE,
        .bDescriptorSubtype     = 1,        /* AS_GENERAL */
        .bTerminalLink          = USBCONFIG_UNIT_TERMINAL_OUT_IQ,
        .bDelay                 = 1,
        .wFormatTag             = 0x0001,   /* PCM */
    },

    /* Format Type Audio */
    .audioFormatType1 = {
        .bLength                = 14,
        .bDescriptorType        = USBD_CS_INTERFACE,
        .bDescriptorSubtype     = 2,        /* FORMAT_TYPE */
        .bFormatType            = 1,        /* TYPE_I */
        .bNrChannels            = 2,
        .bSubFrameSize          = 4,
        .bBitResolution         = 32,
        .bSamFreqType           = 2,        /* 2 frequencies follow */
    },
    .audioSampleFrequencies1 = {
        { LE24(48000UL), },
        { LE24(96000UL), },
    },

    /* Audio streaming IN endpoint */
    .audioEndpoint1 = {
        .bLength                = USBD_SIZE_ENDPOINT_ISO,
        .bDescriptorType        = USBD_ENDPOINT,
        .bEndpointAddress       = USBCONFIG_STREAMING_EP1,
        .bmAttributes           = 0x05,     /* isochronous, async, data */
        .wMaxPacketSize         = USBCONFIG_ISOC_SIZE1_32,
        .bInterval              = 1,
        .bRefresh               = 0,
        .bSyncAddress           = 0,
    },

    /* Class specific audio endpoint */
    .audioDataEndpoint1 = {
        .bLength                = 7,
        .bDescriptorType        = USBD_CS_ENDPOINT,
        .bDescriptorSubtype     = 1,        /* EP_GENERAL */
        .bmAttributes           = 1,        /* Sample rate control */
        .bLockDelayUnits        = 0,
        .wLockDelay             = 0,
    },

    /* Standard AudioStreaming Interface 1 (alternate setting #2) */
    .audioStreamingInterface_1_2 = {
        .bLength                = USBD_SIZE_INTERFACE,
        .bDescriptorType        = USBD_INTERFACE,
        .bInterfaceNumber       = USBCONFIG_INTERFACE_AUDIO_STREAMING_1,
        .bAlternateSetting      = 2,
        .bNumEndpoints          = 1,
        .bInterfaceClass        = USBC_AUDIO,
        .bInterfaceSubClass     = USBCS_AUDIO_AUDIOSTREAMING,
        .bInterfaceProtocol     = 0,
        .iInterface             = USBSTR_STREAM1,
    },

    /* Audio Stream Audio Class */
    .audioClassStreamingInterface2 = {
        .bLength                = 7,
        .bDescriptorType        = USBD_CS_INTERFACE,
        .bDescriptorSubtype     = 1,        /* AS_GENERAL */
        .bTerminalLink          = USBCONFIG_UNIT_TERMINAL_OUT_IQ,
        .bDelay                 = 1,
        .wFormatTag             = 0x0001,   /* PCM */
    },

    /* Format Type Audio */
    .audioFormatType2 = {
#if BOARD_TYPE == BOARD_ROLF2
        .bLength                = 17,
#else
        .bLength                = 14,
#endif
        .bDescriptorType        = USBD_CS_INTERFACE,
        .bDescriptorSubtype     = 2,        /* FORMAT_TYPE */
        .bFormatType            = 1,        /* TYPE_I */
        .bNrChannels            = 2,
        .bSubFrameSize          = 2,
        .bBitResolution         = 16,
#if BOARD_TYPE == BOARD_ROLF2
        .bSamFreqType           = 3,        /* 3 frequencies follow */
#else
        .bSamFreqType           = 2,        /* 2 frequencies follow */
#endif
    },
    .audioSampleFrequencies2 = {
        { LE24(48000UL), },
        { LE24(96000UL), },
#if BOARD_TYPE == BOARD_ROLF2
        { LE24(192000UL), },
#endif
    },

    /* Audio streaming IN endpoint */
    .audioEndpoint2 = {
        .bLength                = USBD_SIZE_ENDPOINT_ISO,
        .bDescriptorType        = USBD_ENDPOINT,
        .bEndpointAddress       = USBCONFIG_STREAMING_EP1,
        .bmAttributes           = 0x05,     /* isochronous, async, data */
        .wMaxPacketSize         = USBCONFIG_ISOC_SIZE1_16,
        .bInterval              = 1,
        .bRefresh               = 0,
        .bSyncAddress           = 0,
    },

    /* Class specific audio endpoint */
    .audioDataEndpoint2 = {
        .bLength                = 7,
        .bDescriptorType        = USBD_CS_ENDPOINT,
        .bDescriptorSubtype     = 1,        /* EP_GENERAL */
        .bmAttributes           = 1,        /* Sample rate control */
        .bLockDelayUnits        = 0,
        .wLockDelay             = 0,
    },

  #if USBCONFIG_DSP
    /* Standard AudioControl Interface */
    .audioDspControlInterface = {
        .bLength                = USBD_SIZE_INTERFACE,
        .bDescriptorType        = USBD_INTERFACE,
        .bInterfaceNumber       = USBCONFIG_INTERFACE_AUDIO_DSP_CONTROL,
        .bAlternateSetting      = 0,
        .bNumEndpoints          = 0,
        .bInterfaceClass        = USBC_AUDIO,
        .bInterfaceSubClass     = USBCS_AUDIO_AUDIOCONTROL,
        .bInterfaceProtocol     = 0,
        .iInterface             = USBSTR_SOUNDCARD2,
    },

    /* Class-Specific Audio Control Interface */
    .audioDspControlInterfaceHeader = {
        .bLength                = USBD_SIZE_AUDIO_ST_HEADER(1),
        .bDescriptorType        = USBD_CS_INTERFACE,
        .bDescriptorSubtype     = USBD_AUDIO_ST_HEADER,
        .bcdADC                 = 0x0100,   /* 1.0 */
        .wTotalLength           = USBD_SIZE_AUDIO_ST_HEADER(1) +
                                  USBD_SIZE_AUDIO_ST_INPUT_TERMINAL +
                                  USBD_SIZE_AUDIO_ST_OUTPUT_TERMINAL +
                                  USBD_SIZE_AUDIO_ST_FEATURE_UNIT(1,1) +
                                  0,
        .bInCollection          = 1,
    },
    .audioDspControlInterfaceHeader_baInterfaceNr = {
        USBCONFIG_INTERFACE_AUDIO_DSP_STREAMING,
    },

    /* Input Terminal (DSP) */
    .audioDspInputTerminal = {
        .bLength                = USBD_SIZE_AUDIO_ST_INPUT_TERMINAL,
        .bDescriptorType        = USBD_CS_INTERFACE,
        .bDescriptorSubtype     = USBD_AUDIO_ST_INPUT_TERMINAL,
        .bTerminalID            = USBCONFIG_UNIT_TERMINAL_IN_DSP,
        .wTerminalType          = 0x0201,   /* Line In */
        .bAssocTerminal         = 0,        /* no association */
        .bNrChannels            = 1,
        .wChannelConfig         = 0x0004,   /* Center C */
        .iChannelNames          = 0,
        .iTerminal              = USBSTR_DEMOD_INPUT_TERMINAL,
    },

    /* Output Terminal DSP */
    .audioDspOutputTerminal = {
        .bLength                = USBD_SIZE_AUDIO_ST_OUTPUT_TERMINAL,
        .bDescriptorType        = USBD_CS_INTERFACE,
        .bDescriptorSubtype     = USBD_AUDIO_ST_OUTPUT_TERMINAL,
        .bTerminalID            = USBCONFIG_UNIT_TERMINAL_OUT_DSP,
        .wTerminalType          = 0x0101,   /* USB Streaming */
        .bAssocTerminal         = 0,        /* no association */
        .bSourceID              = USBCONFIG_UNIT_FEATURE_DSP,
        .iTerminal              = 0,
    },

    /* Feature Unit (DSP) */
    .audioDspFeatureUnit = {
        .bLength                = USBD_SIZE_AUDIO_ST_FEATURE_UNIT(1,1),
        .bDescriptorType        = USBD_CS_INTERFACE,
        .bDescriptorSubtype     = USBD_AUDIO_ST_FEATURE_UNIT,
        .bUnitID                = USBCONFIG_UNIT_FEATURE_DSP,
        .bSourceID              = USBCONFIG_UNIT_TERMINAL_IN_DSP,
        .bControlSize           = 1,        /* n = 1 Byte/channel */
    },
    .audioDspFeatureUnit_bmaControls = {
            0x43,               /* bmaControls(0)       Master: Mute, Volume, AGC */
            0x00,               /* bmaControls(1)                                 */
    },
    .audioDspFeatureUnit_iFeature = 0,

    /* Standard AudioStreaming Interface 2 (alternate setting #0) */
    .audioDspStreamingInterface_0 = {
        .bLength                = USBD_SIZE_INTERFACE,
        .bDescriptorType        = USBD_INTERFACE,
        .bInterfaceNumber       = USBCONFIG_INTERFACE_AUDIO_DSP_STREAMING,
        .bAlternateSetting      = 0,
        .bNumEndpoints          = 0,
        .bInterfaceClass        = USBC_AUDIO,
        .bInterfaceSubClass     = USBCS_AUDIO_AUDIOSTREAMING,
        .bInterfaceProtocol     = 0,
        .iInterface             = USBSTR_STREAM2,
    },

    /* Standard AudioStreaming Interface 2 (alternate setting #1) */
    .audioDspStreamingInterface_1 = {
        .bLength                = USBD_SIZE_INTERFACE,
        .bDescriptorType        = USBD_INTERFACE,
        .bInterfaceNumber       = USBCONFIG_INTERFACE_AUDIO_DSP_STREAMING,
        .bAlternateSetting      = 1,
        .bNumEndpoints          = 1,
        .bInterfaceClass        = USBC_AUDIO,
        .bInterfaceSubClass     = USBCS_AUDIO_AUDIOSTREAMING,
        .bInterfaceProtocol     = 0,
        .iInterface             = USBSTR_STREAM2,
    },

    /* Audio Stream Audio Class */
    .audioDspClassStreamingInterface = {
        .bLength                = 7,
        .bDescriptorType        = USBD_CS_INTERFACE,
        .bDescriptorSubtype     = 1,        /* AS_GENERAL */
        .bTerminalLink          = USBCONFIG_UNIT_TERMINAL_OUT_DSP,
        .bDelay                 = 1,
        .wFormatTag             = 0x0001,   /* PCM */
    },

    /* Format Type Audio */
    .audioDspFormatType = {
        .bLength                = 11,
        .bDescriptorType        = USBD_CS_INTERFACE,
        .bDescriptorSubtype     = 2,        /* FORMAT_TYPE */
        .bFormatType            = 1,        /* TYPE_I */
        .bNrChannels            = 1,
        .bSubFrameSize          = 2,
        .bBitResolution         = 16,
        .bSamFreqType           = 1,        /* 1 frequency follows */
    },
    .audioDspSampleFrequencies = {
        { LE24(12000ul), },
    },

    /* Audio streaming IN endpoint */
    .audioDspEndpoint = {
        .bLength                = USBD_SIZE_ENDPOINT_ISO,
        .bDescriptorType        = USBD_ENDPOINT,
        .bEndpointAddress       = USBCONFIG_STREAMING_EP2,
        .bmAttributes           = 0x05,     /* isochronous, async, data */
        .wMaxPacketSize         = USBCONFIG_ISOC_SIZE2,
        .bInterval              = 1,
        .bRefresh               = 0,
        .bSyncAddress           = 0,
    },

    /* Class specific audio endpoint */
    .audioDspDataEndpoint = {
        .bLength                = 7,
        .bDescriptorType        = USBD_CS_ENDPOINT,
        .bDescriptorSubtype     = 1,        /* EP_GENERAL */
        .bmAttributes           = 1,        /* Sample rate control */
        .bLockDelayUnits        = 0,
        .wLockDelay             = 0,
    },
  #endif
#endif  /* #if USBCONFIG_SOUNDCARD */
};


/** Array with all configurations (we have only one). */
static const USB_ConfigurationArrayElement fifisdr_configurations [USBCONFIG_NUM_CONFIGURATIONS] = {
    {sizeof(fifisdr_configuration1), &fifisdr_configuration1, },
};


/** Complete collection of all FiFi-SDR descriptors */
const USB_DescriptorList fifisdr_descriptors = {
    .device                 = &fifisdr_deviceDescriptor,
    .numConfigurations      = USBCONFIG_NUM_CONFIGURATIONS,
    .callbackGetString      = fifisdr_getString,
    .configurations         = fifisdr_configurations,
};



/** Do initialization for the USB descriptors. */  //TODO reduce stack usage
void fifisdr_initDescriptors (void)
{
    uint32_t command[5];
    uint32_t result[5];
    typedef void (*IAP)(uint32_t [], uint32_t[]);
    uint8_t *p;
    uint8_t i, j;
    uint8_t digit;


    /* Read the serial number of the LPC, and transform it into a USB string descriptor.
     * Issue IAP command 58: Read device serial number
     */
    command[0] = 58;
    ((IAP)0x1FFF1FF1)(command, result);

    /* Make a string out of it */
    p = (uint8_t *)&str_iSerialNumber;
    *p++ = sizeof(str_iSerialNumber);
    *p++ = USBD_STRING;
    for (i = 0; i < 4; i++) {
        for (j = 0; j < 8; j++) {
            digit = (result[4-i] >> 28) & 0x0F;
            if (digit < 10) {
                *p++ = '0' + digit;
            }
            else {
                *p++ = '7' + digit;
            }
            *p++ = 0;
            result[4-i] <<= 4;
        }
    }
}




