/* Copyright (c) 2011-2012, DF9DQ
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this list of conditions
 * and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions
 * and the following disclaimer in the documentation and/or other materials provided with the
 * distribution.
 * Neither the name of the author nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __USBAUDIO_H
#define __USBAUDIO_H

#include <stdbool.h>

#include "usbclass.h"


/* USB Class Codes and applicable sub classes and protocols */
#define USBC_AUDIO                                      (0x01)
#define   USBCS_AUDIO_SUBCLASS_UNDEFINED                    (0x00)
#define   USBCS_AUDIO_AUDIOCONTROL                          (0x01)
#define   USBCS_AUDIO_AUDIOSTREAMING                        (0x02)
#define   USBCS_AUDIO_MIDISTREAMING                         (0x03)
#define     USBCP_AUDIO_PR_PROTOCOL_UNDEFINED                   (0x00)


/* Class-specific Descriptors (and their length) */
#define USBD_AUDIO_CS_UNDEFINED                         (0x20)
#define USBD_SIZE_AUDIO_CS_UNDEFINED                    (5)
#define USBD_AUDIO_CS_DEVICE                            (0x21)
#define USBD_AUDIO_CS_CONFIGURATION                     (0x22)
#define USBD_AUDIO_CS_STRING                            (0x23)
#define USBD_AUDIO_CS_INTERFACE                         (0x24)
#define USBD_AUDIO_CS_ENDPOINT                          (0x25)


/* Audio Class Control Selectors */
#define USBAC_CS_FU_MUTE_CONTROL                0x01
#define USBAC_CS_FU_VOLUME_CONTROL              0x02
#define USBAC_CS_FU_BASS_CONTROL                0x03
#define USBAC_CS_FU_MID_CONTROL                 0x04
#define USBAC_CS_FU_TREBLE_CONTROL              0x05
#define USBAC_CS_FU_GRAPHIC_EQUALIZER_CONTROL   0x06
#define USBAC_CS_FU_AUTOMATIC_GAIN_CONTROL      0x07
#define USBAC_CS_FU_DELAY_CONTROL               0x08
#define USBAC_CS_FU_BASS_BOOST_CONTROL          0x09
#define USBAC_CS_FU_LOUDNESS_CONTROL            0x0A


/* Processing Unit Process Type */
#define USBAC_CS_PU_UPDOWNMIX_PROCESS           0x01
#define USBAC_CS_PU_DOLBY_PROLOGIC_PROCESS      0x02
#define USBAC_CS_PU_3D_STEREO_EXTENDER_PROCESS  0x03
#define USBAC_CS_PU_REVERBERATION_PROCESS       0x04
#define USBAC_CS_PU_CHORUS_PROCESS              0x05
#define USBAC_CS_PU_DYN_RANGE_COMP_PROCESS      0x06


#define USBAC_CS_EP_SAMPLING_FREQ_CONTROL       0x01
#define USBAC_CS_EP_PITCH_CONTROL               0x02



#define SET_CUR                                 0x01
#define GET_CUR                                 0x81
#define SET_MIN                                 0x02
#define GET_MIN                                 0x82
#define SET_MAX                                 0x03
#define GET_MAX                                 0x83
#define SET_RES                                 0x04
#define GET_RES                                 0x84
#define SET_MEM                                 0x05
#define GET_MEM                                 0x85
#define GET_STAT                                0xFF



/** Minimum, maximum and resolution for a control in a unit. */
typedef struct USBAUDIO_ControlRange {
    uint8_t     unit_id;
    uint8_t     control_id;
    int16_t     min;
    int16_t     max;
    int16_t     current;
    uint16_t    resolution;
    uint8_t     length;
    LPCLIB_Callback callback;
} USBAUDIO_ControlRange;


/** USB Class-Specific AC Interface descriptor. */
typedef __PACKED(struct USBAUDIO_ControlInterfaceDescriptor {
    uint8_t     bLength;
    uint8_t     bDescriptorType;
    uint8_t     bDescriptorSubtype;
    uint16_t    bcdADC;
    uint16_t    wTotalLength;
    uint8_t     bInCollection;
    /*uint8_t     baInterfaceNr[];*/ /* Must be declared separately! */
}) USBAUDIO_ControlInterfaceDescriptor;


/** USB Audio Input Terminal Descriptor. */
typedef __PACKED(struct USBAUDIO_InputTerminalDescriptor {
    uint8_t     bLength;
    uint8_t     bDescriptorType;
    uint8_t     bDescriptorSubtype;
    uint8_t     bTerminalID;
    uint16_t    wTerminalType;
    uint8_t     bAssocTerminal;
    uint8_t     bNrChannels;
    uint16_t    wChannelConfig;
    uint8_t     iChannelNames;
    uint8_t     iTerminal;
}) USBAUDIO_InputTerminalDescriptor;


/** USB Audio Output Terminal Descriptor. */
typedef __PACKED(struct USBAUDIO_OutputTerminalDescriptor {
    uint8_t     bLength;
    uint8_t     bDescriptorType;
    uint8_t     bDescriptorSubtype;
    uint8_t     bTerminalID;
    uint16_t    wTerminalType;
    uint8_t     bAssocTerminal;
    uint8_t     bSourceID;
    uint8_t     iTerminal;
}) USBAUDIO_OutputTerminalDescriptor;


/** USB Feature Unit Descriptor. */
typedef __PACKED(struct USBAUDIO_FeatureUnitDescriptor {
    uint8_t     bLength;
    uint8_t     bDescriptorType;
    uint8_t     bDescriptorSubtype;
    uint8_t     bUnitID;
    uint8_t     bSourceID;
    uint8_t     bControlSize;
    /*uint8_t       bmaControls[];*/ /* Must be declared separately! */
    /*uint8_t       iFeature;*/ /* Must be declared separately! */
}) USBAUDIO_FeatureUnitDescriptor;


/** USB Processing Unit Descriptor. */
typedef __PACKED(struct USBAUDIO_ProcessingUnitDescriptor {
    uint8_t     bLength;
    uint8_t     bDescriptorType;
    uint8_t     bDescriptorSubtype;
    uint8_t     bUnitID;
    uint16_t    wProcessType;
    uint8_t     bNrInPins;
    uint8_t     bSourceID;
    uint8_t     bNrChannels;
    uint16_t    wChannelConfig;
    uint8_t     iChannelNames;
    uint8_t     bControlSize;
    /*uint8_t       bmControls[];*/ /* Must be declared separately! */
    /*uint8_t       iProcessing;*/ /* Must be declared separately! */
}) USBAUDIO_ProcessingUnitDescriptor;


/** USB Selector Unit Descriptor. */
typedef __PACKED(struct USBAUDIO_SelectorUnitDescriptor {
    uint8_t     bLength;
    uint8_t     bDescriptorType;
    uint8_t     bDescriptorSubtype;
    uint8_t     bUnitID;
    uint8_t     bNrInPins;
    /*uint8_t       baSourceID[];*/ /* Must be declared separately! */
    /*uint8_t       iSelector;*/ /* Must be declared separately! */
}) USBAUDIO_SelectorUnitDescriptor;


/** USB Microphone Type I Format Type Descriptor. */
typedef __PACKED(struct USBAUDIO_MicrophoneType1FormatTypeDescriptor {
    uint8_t     bLength;
    uint8_t     bDescriptorType;
    uint8_t     bDescriptorSubtype;
    uint8_t     bFormatType;
    uint8_t     bNrChannels;
    uint8_t     bSubFrameSize;
    uint8_t     bBitResolution;
    uint8_t     bSamFreqType;
    /*uint8_t       tSamFreq[3][];*/ /* Must be declared separately! */
}) USBAUDIO_MicrophoneType1FormatTypeDescriptor;


/** USB Audio Streaming Interface Descriptor. */
typedef __PACKED(struct USBAUDIO_StreamingInterfaceDescriptor {
    uint8_t     bLength;
    uint8_t     bDescriptorType;
    uint8_t     bDescriptorSubtype;
    uint8_t     bTerminalLink;
    uint8_t     bDelay;
    uint16_t    wFormatTag;
}) USBAUDIO_StreamingInterfaceDescriptor;


/** USB Audio Data Endpoint Descriptor. */
typedef __PACKED(struct USBAUDIO_IsoDataEndpointDescriptor {
    uint8_t     bLength;
    uint8_t     bDescriptorType;
    uint8_t     bDescriptorSubtype;
    uint8_t     bmAttributes;
    uint8_t     bLockDelayUnits;
    uint16_t    wLockDelay;
}) USBAUDIO_IsoDataEndpointDescriptor;


/**********************************************************************************************/


/** Audio control interface */
typedef struct {
    uint8_t     interfaceNumber;
    bool        usesInterrupt;
    uint8_t     endpointNumberInterrupt;
    uint8_t     numControls;
    USBAUDIO_ControlRange *controls;
} USBAUDIO_ControlInterface;

/** One alternative for an audio streaming interface. */
typedef struct {
    uint8_t     numEndpoints;
    uint8_t     linkedTerminal;
    uint8_t     endpointNumber;
    uint16_t    packetSize;
    uint16_t    supportedControls;
    uint8_t     numSamplingFrequencies;
    const uint32_t  *samplingFrequencies;
} USBAUDIO_StreamingInterfaceParams;

/** Complete audio streaming interface. */
typedef struct {
    uint8_t     interfaceNumber;
    uint8_t     numAltSettings;
    uint8_t     activeSetting;
    uint32_t    currentSamplerate;
    bool        isMuted;
    const USBAUDIO_StreamingInterfaceParams *params;
} USBAUDIO_StreamingInterface;


/** Complete description of an Audio Class function. */
typedef struct {
    const USBAUDIO_ControlInterface *controlInterface;
    uint8_t     numStreamingInterfaces;
    USBAUDIO_StreamingInterface *streamingInterfaces;
    LPCLIB_Callback callback;
} USBAUDIO_FunctionDeclaration;



/**********************************************************************************************/

/** Handle for a USB audio class instance. */
typedef struct USBAUDIO_Context *USBAUDIO_Handle;


/** Opcodes to specify the configuration command in a call to \ref USBAUDIO_ioctl. */
typedef enum USBAUDIO_Opcode {
    USBAUDIO_OPCODE_xxx,                    /**< Config action: ... */
} USBAUDIO_Opcode;



/** Descriptor to specify the configuration in a call to \ref USBAUDIO_ioctl. */
typedef struct USBAUDIO_Config {
    USBAUDIO_Opcode opcode;                 /**< Config action opcode */
    LPCLIB_Switch continued;                /**< Set if further config struct's follow in an array */

    union {
        int dummy;
    };
} USBAUDIO_Config;


/* Event opcodes for audio class */
typedef enum USBAUDIO_CallbackEvent {
    USBAUDIO_EVENT_INTERFACE_CHANGE = 0,    /**< Interface selection (alternate setting) */
    USBAUDIO_EVENT_SET_CONTROL,             /**< Set value of a control */
    USBAUDIO_EVENT_GET_CONTROL,             /**< Get value of a control */
    USBAUDIO_EVENT_ENDPOINT,                /**< Set an endpoint parameter */
} USBAUDIO_CallbackEvent;




/** Open a USB Audio Class instance.
 *
 *  \param[in] pFunction Audio class function definition
 *  \param[out] pHandle Audio class handle
 *  \retval LPCLIB_SUCCESS Success. \ref handle contains a valid handle.
 *  \retval LPCLIB_BUSY Failure
 */
LPCLIB_Result USBAUDIO_open (const USBAUDIO_FunctionDeclaration *pFunction, USBAUDIO_Handle *pHandle);


/** Configure the USB audio class
 *
 *  \param[in] handle USB audio class handle.
 *  \param[in] pConfig Pointer to a configuration descriptor
 */
void USBAUDIO_ioctl (USBAUDIO_Handle handle, const USBAUDIO_Config *pConfig);


/** USB audio class request handler.
 *
 *  \param[in] device USB device handle
 *  \param[in] pClass Audio class handle
 *  \param[in] pSetup SETUP packet
 *  \param[in] pBuffer Control endpoint buffer (SETUP data and/or response)
 *  \retval LPCLIB_SUCCESS Request addressed to this class.
 *                         Request valid and handled.
 *  \retval LPCLIB_ILLEGAL_PARAMETER Request addressed to this class.
 *                         Request invalid.
 *  \retval LPCLIB_UNDEFINED Request not handled by this class.
 */
LPCLIB_Result USBAUDIO_handleRequests (USBDEV_Handle device,
                                       const struct USB_Class *pClass,
                                       USB_SetupPacket *pSetup,
                                       USB_Buffer *pBuffer);

#endif
